
import json
import httpx
import datetime

from enum import Enum
from pydantic import validator
from pydantic.types import ByteSize
from async_openai.types import BaseModel, lazyproperty
from typing import List, Any, Optional, Type, Dict, Union, Iterator, cast, Callable
from async_openai.utils import logger
from async_openai.utils.helpers import is_coro_func

from async_openai.schemas.types.exceptions import error_handler

from fileio import File, FileType

__all__ = [
    'BaseEndpoint',
    'Permission',
    'Usage',
    'FilePurpose',
    'FileRequest',
    'FileData',
    'EventResult',
    'EngineResult',
    'BaseResult',
]

class Method(str, Enum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    PATCH = "PATCH"
    HEAD = "HEAD"

class BaseEndpoint(BaseModel):
    method: Method
    url: str
    data: Optional[Union[Dict, Callable, Any]]
    params: Optional[Union[Dict, Callable, Any]]
    files: Optional[Union[Dict, Callable, Any]]
    sync_files: Optional[Union[Dict, Callable, Any]]

    async def async_get_params(
        self,
        **kwargs
    ):
        params = {
            "method": self.method.value,
            "url": self.url,
        }
        if self.data:
            if isinstance(self.data, dict):
                params["json"] = self.data
            elif isinstance(self.data, bytes):
                params["data"] = self.data
            elif is_coro_func(self.data):
                params["json"] = await self.data(**kwargs)
            elif callable(self.data):
                params["json"] = self.data(**kwargs)
            else:
                params["json"] = self.data
        if self.params:
            if isinstance(self.params, dict):
                params["params"] = self.params
            elif is_coro_func(self.params):
                params["params"] = await self.params(**kwargs)
            elif callable(self.params):
                params["params"] = self.params(**kwargs)
            else:
                params["params"] = self.params
        if self.files:
            if is_coro_func(self.files):
                params["files"] = await self.files(**kwargs)
            elif callable(self.files):
                params["files"] = self.files(**kwargs)
            else:
                params["files"] = self.files
        elif self.sync_files:
            if is_coro_func(self.sync_files):
                params["files"] = await self.sync_files(**kwargs)
            elif callable(self.sync_files):
                params["files"] = self.sync_files(**kwargs)
            else:
                params["files"] = self.sync_files
        return params

    def get_params(
        self,
        **kwargs
    ):
        params = {
            "method": self.method.value,
            "url": self.url,
        }
        if self.data:
            if isinstance(self.data, dict) or not isinstance(self.data, bytes) and not callable(self.data):
                params["json"] = self.data
            elif isinstance(self.data, bytes):
                params["data"] = self.data
            else:
                params["json"] = self.data(**kwargs)
        if self.params:
            if isinstance(self.params, dict) or not callable(self.params):
                params["params"] = self.params
            else:
                params["params"] = self.params(**kwargs)
        if self.sync_files:
            if callable(self.sync_files):
                params["files"] = self.sync_files(**kwargs)
            else:
                params["files"] = self.sync_files
        elif self.files:
            params["files"] = self.files(**kwargs) if callable(self.files) else self.files
        return params


class Permission(BaseModel):
    id: str
    object: str
    created: datetime.datetime
    allow_create_engine: bool
    allow_sampling: bool
    allow_logprobs: bool
    allow_search_indices: bool
    allow_view: bool
    allow_fine_tuning: bool
    organization: str
    group: Optional[str]
    is_blocking: bool

    @property
    def since_seconds(self):
        return (datetime.datetime.now(datetime.timezone.utc) - self.created).total_seconds()

class Usage(BaseModel):
    prompt_tokens: Optional[int]
    completion_tokens: Optional[int]
    total_tokens: Optional[int]


class FilePurpose(str, Enum):
    """
    File Purpose
    """

    finetune = "fine-tune"
    fine_tune = "fine-tune"
    train = "fine-tune-train"
    search = "search"

    @classmethod
    def parse_str(cls, value: str):
        if "train" in value:
            return cls.train
        elif "finetune" in value:
            return cls.finetune
        elif "fine-tune" in value:
            return cls.fine_tune
        elif "search" in value:
            return cls.search
        raise ValueError(f"Cannot convert {value} to FilePurpose")


class FileRequest(BaseModel):
    file_id: Optional[str]
    file: Optional[Union[str, FileType, Any]]
    filename: Optional[str] = None
    purpose: FilePurpose = FilePurpose.fine_tune
    model: Optional[str]

    @validator("purpose")
    def validate_purpose(cls, value):
        return FilePurpose.parse_str(value) if isinstance(value, str) else value
    
    def get_file_request_params(self, **kwargs) -> Dict:
        """
        Transforms the data to the req params
        """
        files = [("purpose", (None, self.purpose.value))]
        if self.purpose == FilePurpose.search and self.model:
            files.append(("model", (None, self.model)))
        if self.file:
            file = File(self.file)
            files.append(
                ("file", (self.filename or file.name, file.read_bytes(), "application/octet-stream"))
            )
        return files
    
    async def async_get_file_request_params(self, **kwargs) -> Dict:
        """
        Transforms the data to the req params
        """
        files = [("purpose", (None, self.purpose.value))]
        if self.purpose == FilePurpose.search and self.model:
            files.append(("model", (None, self.model)))
        if self.file:
            file = File(self.file)
            files.append(
                ("file", (self.filename or file.name, await file.async_read_bytes(), "application/octet-stream"))
            )
        return files

    @lazyproperty
    def list_files_endpoint(self) -> BaseEndpoint:
        return BaseEndpoint(
            method = Method.GET,
            url = '/files',
        )
    
    @lazyproperty
    def upload_file_endpoint(self) -> BaseEndpoint:
        return BaseEndpoint(
            method = Method.POST,
            url = '/files',
            files = self.async_get_file_request_params,
            sync_files = self.get_file_request_params
        )

    @property
    def delete_file_endpoint(self) -> BaseEndpoint:
        return BaseEndpoint(
            method = Method.DELETE,
            url = f'/files/{self.file_id}'
        )

    @property
    def retrieve_file_endpoint(self) -> BaseEndpoint:
        return BaseEndpoint(
            method = Method.GET,
            url = f'/files/{self.file_id}'
        )

    @property
    def retrieve_file_content_endpoint(self) -> BaseEndpoint:
        return BaseEndpoint(
            method = Method.GET,
            url = f'/files/{self.file_id}/content'
        )


class FileData(BaseModel):
    id: str
    object: Optional[str] = 'file'
    bytes: Optional[ByteSize]
    created_at: Optional[datetime.datetime]
    filename: Optional[str]
    purpose: Optional[FilePurpose] = FilePurpose.fine_tune

    @validator("created_at")
    def validate_created_at(cls, value):
        return datetime.datetime.fromtimestamp(value, datetime.timezone.utc) if value else value
    

class EventResult(BaseModel):
    object: Optional[str]
    created_at: Optional[datetime.datetime]
    level: Optional[str]
    message: Optional[str]

    @property
    def since_seconds(self) -> int:
        if self.created_at is None: return -1
        return (datetime.datetime.now(datetime.timezone.utc) - self.created_at).total_seconds()


class EngineResult(BaseModel):
    id: Optional[str]
    object: Optional[str]
    owner: Optional[str]
    ready: Optional[bool]


class BaseResult(BaseModel):
    id: Optional[str]
    object: Optional[str]
    created: Optional[datetime.datetime]
    updated: Optional[datetime.datetime]
    choices: Optional[List[Any]]
    data: Optional[Union[Any, List[Any]]] # Used for single retrieval if not in list.
    events: Optional[List[Any]]

    usage: Optional[Usage] = None
    model: Optional[str] = None

    _request: Optional[Type[BaseModel]] = None
    _raw_request: Optional[httpx.Request] = None
    _raw_response: Optional[httpx.Response] = None

    _choice_model: Optional[Type[BaseModel]] = None
    _data_model: Optional[Type[BaseModel]] = None
    _event_model: Optional[Type[BaseModel]] = None

    @property
    def metadata_fields(self):
        # Default Metadata fields
        return [
            'id',
            'object',
            'created',
            # 'choices',
            'model',
            'usage',
        ]

    @property
    def since_seconds(self) -> int:
        return (datetime.datetime.now(datetime.timezone.utc) - self.created).total_seconds() \
            if self.created else -1

    @property
    def request_headers(self) -> Dict[str, str]:
        return self._raw_request.headers if self._raw_request else {}
    
    @property
    def response_headers(self) -> Dict[str, str]:
        return self._raw_response.headers if self._raw_response else {}

    @property
    def is_stream(self):
        return "text/event-stream" in self.response_headers.get("content-type", "")
    
    @property
    def request_id(self) -> Optional[str]:
        return self.response_headers.get("request-id", self.response_headers.get("x-request-id"))
    
    @property
    def organization(self) -> str:
        return self.response_headers.get("openai-organization")

    @property
    def response_ms(self) -> Optional[int]:
        h = self.response_headers.get("openai-processing-ms")
        return None if h is None else round(float(h))
    
    @property
    def response_timestamp(self) -> Optional[datetime.datetime]:
        dt = self.response_headers.get("date")
        return None if dt is None else datetime.datetime.strptime(dt, '%a, %d %b %Y %H:%M:%S GMT')

    @property
    def openai_version(self) -> Optional[str]:
        return self.response_headers.get("openai-version")
    
    @property
    def openai_model(self) -> Optional[str]:
        return self.response_headers.get("openai-model")
    
    def __len__(self):
        return len(self.choices) if self.choices else 0
    
    def __iter__(self):
        return iter(self.choices) if self.choices else iter([])
    
    def __getitem__(self, idx: int) -> Union[Type[BaseModel], Dict, Any]:
        return self.choices[idx] if self.choices else None

    def parse_stream(self, response: httpx.Response) -> Iterator[Dict]:
        for line in response.iter_lines():
            if not line: continue
            if "data: [DONE]" in line:
                # return here will cause GeneratorExit exception in urllib3
                # and it will close http connection with TCP Reset
                continue
            if line.startswith("data: "):
                line = line[len("data: ") :]
            if not line.strip(): continue
            try:
                yield json.loads(line)
            except Exception as e:
                logger.error(f'Error: {line}: {e}')
    
    def set_obj_meta(
        self,
        item: Dict
    ):
        """
        Sets the appropriate metadata fields
        """
        for key in self.metadata_fields:
            if key in item:
                if key == 'usage':
                    self.usage = Usage(**item['usage'])
                elif key in {'created', 'created_at'}:
                    self.created = datetime.datetime.fromtimestamp(item[key], datetime.timezone.utc)
                elif key in {'updated', 'updated_at'}:
                    self.updated = datetime.datetime.fromtimestamp(item[key], datetime.timezone.utc)
                elif key in {'result_files', 'training_files', 'validation_files'}:
                    vals = [
                        FileData.parse_obj(val) for val in item[key]
                    ]
                    setattr(self, key, vals)

                else: setattr(self, key, cast(type(getattr(self, key)), item[key]))
        
    def handle_data_item(
        self,
        item: Dict,
    ):
        """
        The handler function for each data item
        """
        if item.get('choices'):
            if self.choices is None: self.choices = []
            choices = item.get('choices', [])
            if self._choice_model:
                choices = [self._choice_model.parse_obj(choice) for choice in choices]
            self.choices.extend(choices)
        elif item.get('data'):
            if self.data is None: self.data = []
            data = item.get('data', [])
            if self._data_model:
                data = [self._data_model.parse_obj(d) for d in data]
            self.data.extend(data)
        elif item.get('events'):
            if self.events is None: self.events = []
            events = item.get('events', [])
            if self._event_model:
                events = [self._event_model.parse_obj(e) for e in events]
            self.events.extend(events)

        else:
            self.data = self._data_model.parse_obj(item) if self._data_model else item


    def handle_response(
        self,
    ):
        """
        The handler function for the response
        """
        if self.is_stream:
            has_set_meta = False
            for item in self.parse_stream(
                self._raw_response
            ):
            
                if "error" in item:
                    # handle error
                    raise error_handler(
                        response = self._raw_response,
                        stream = True,
                        data = item,
                    )
                if not has_set_meta:
                    # Set the initial params that are created from the first one.
                    self.set_obj_meta(item)
                    has_set_meta = True
                self.handle_data_item(item)
        else:
            item = self._raw_response.json()
            self.set_obj_meta(item)
            self.handle_data_item(item)

    def parse_result(
        self,
        raw_response: Optional[httpx.Response] = None,
        raw_request: Optional[httpx.Request] = None,
    ):
        if raw_response: self._raw_response = raw_response
        if raw_request: self._raw_request = raw_request
        self.handle_response()





from typing import List, Any, Optional, Union, Dict, Type
from async_openai.schemas.types.base import BaseResult, FileData, FileRequest

__all__ = [
    'FileRequest',
    'FileData',
    'FileResult',
]

class FileResult(BaseResult):
    data: Optional[Union[FileData, List[FileData]]]
    _data_model: Optional[Type[FileData]] = FileData
    _request: Optional[FileRequest] = None

    @property
    def metadata_fields(self):
        return [
            'created',
            'object',
            # 'data',
        ]









from async_openai.types import BaseModel, lazyproperty
from typing import List, Any, Optional, Union, Type
from async_openai.schemas.types.base import Permission, BaseResult, Method, BaseEndpoint

__all__ = [
    'ModelData',
    'ModelRequest',
    'ModelResult',
]

class ModelData(BaseModel):
    id: str
    owned_by: str
    permission: List[Permission] = []
    root: str
    parent: Optional[str]
    object: str = 'model'

    @property
    def openai_id(self):
        return self.id

class ModelRequest(BaseModel):
    model: Optional[str]

    @lazyproperty
    def list_models_endpoint(self) -> BaseEndpoint:
        return BaseEndpoint(
            method = Method.GET,
            url ='/models',
        )

    @lazyproperty
    def retrieve_model_endpoint(self) -> BaseEndpoint:
        return BaseEndpoint(
            method = Method.GET,
            url = f'/models/{self.model}'
        )
    

class ModelResult(BaseResult):
    data: Optional[Union[ModelData, List[ModelData]]]
    _data_model: Optional[Type[ModelData]] = ModelData
    _request: Optional[ModelRequest] = None

    @property
    def metadata_fields(self):
        return [
            'object',
            # 'data',
        ]






import datetime
from enum import Enum

from pydantic import validator
from typing import List, Any, Optional, Union, Dict, Type
from async_openai.types import BaseModel, lazyproperty
from async_openai.schemas.types.base import BaseResult, EventResult, Method, BaseEndpoint
from async_openai.schemas.types.files import FileRequest, FileData
from async_openai.utils import logger

__all__ = [
    'FinetuneModels',
    'FinetuneRequest',
    'FinetuneData',
    'FinetuneResult',
]

class FinetuneModels(str, Enum):
    """
    Supported finetune models.
    """
    ada = "ada"
    babbage = "babbage"
    curie = "curie"
    davici = "davici"


class FinetuneRequest(BaseModel):
    """Finetune request model"""

    fine_tune_id: Optional[str]
    training_file: Optional[str] # file-xxxx
    validation_file: Optional[str] # file-xxxx

    model: Optional[Union[str, FinetuneModels]] = FinetuneModels.curie
    n_epochs: Optional[int] = 4
    batch_size: Optional[int] = None
    learning_rate_multiplier: Optional[float] = None
    prompt_loss_weight: Optional[float] = 0.01
    compute_classification_metrics: Optional[bool] = False
    classification_n_classes: Optional[int] = None
    classification_betas: Optional[List[float]] = None
    suffix: Optional[str] = None
    stream: bool = False

    @validator("batch_size")
    def validate_batch_size(cls, value):
        """
        Batch Size is capped at 256
        """
        return value if value is None else min(int(value), 256)

    @validator("learning_rate_multiplier")
    def validate_learning_rate_multiplier(cls, value):
        """
        The learning rate multiplier to use for training. 
        The fine-tuning learning rate is the original learning rate used for 
        pretraining multiplied by this value.

        By default, the learning rate multiplier is the 0.05, 0.1, or 0.2 
        depending on final batch_size 
        (larger learning rates tend to perform better with larger batch sizes). 
        We recommend experimenting with values in the range 0.02 to 0.2 to see 
        what produces the best results.
        """
        return value if value is None else min(float(value), 2.0)
    
    @validator("model")
    def validate_model(cls, value):
        """
        Supported models:
        - ada
        - babbage
        - curie
        - davici
        """
        if value is None:
            return FinetuneModels.curie
        if isinstance(value, str) and value in FinetuneModels.__members__:
            return FinetuneModels(value)
        logger.warning(f'{value} is not a validated model. Skipping check.')
        return value
    
    @lazyproperty
    def create_finetune_endpoint(self) -> BaseEndpoint:
        return BaseEndpoint(
            method = Method.POST,
            url = '/fine-tunes',
            data = self.dict(
                exclude_none = True,
                exclude = {'stream'},
            )
        )
    
    @lazyproperty
    def list_finetune_endpoint(self) -> BaseEndpoint:
        return BaseEndpoint(
            method = Method.GET,
            url = '/fine-tunes'
        )

    @property
    def retrieve_finetune_endpoint(self) -> BaseEndpoint:
        return BaseEndpoint(
            method = Method.GET,
            url = f'/fine-tunes/{self.fine_tune_id}'
        )
    
    @property
    def cancel_finetune_endpoint(self) -> BaseEndpoint:
        return BaseEndpoint(
            method = Method.POST,
            url = f'/fine-tunes/{self.fine_tune_id}/cancel'
        )
    
    @property
    def list_finetune_events_endpoint(self) -> BaseEndpoint:
        return BaseEndpoint(
            method = Method.GET,
            url = f'/fine-tunes/{self.fine_tune_id}/events',
            params = {
                'stream': self.stream,
            }
        )
    
    @property
    def delete_finetune_model_endpoint(self) -> BaseEndpoint:
        return BaseEndpoint(
            method = Method.DELETE,
            url = f'/models/{self.model}'
        )

    
class FinetuneData(BaseResult):
    id: Optional[str]
    object: Optional[str] = 'fine-tune'
    model: Optional[str]
    created_at: Optional[datetime.datetime]
    fine_tuned_model: Optional[str]
    hyperparams: Optional[Dict[str, Any]]
    organization_id: Optional[str]
    result_files: Optional[List[FileData]]
    status: Optional[str]
    training_files: Optional[List[FileData]]
    validation_files: Optional[List[FileData]]
    updated_at: Optional[datetime.datetime]

    @validator("created_at")
    def validate_created_at(cls, value):
        return datetime.datetime.fromtimestamp(value, datetime.timezone.utc) if value else value
    
    @validator("updated_at")
    def validate_updated_at(cls, value):
        return datetime.datetime.fromtimestamp(value, datetime.timezone.utc) if value else value
    


class FinetuneResult(BaseResult):
    events: Optional[List[EventResult]]
    data: Optional[Union[FinetuneData, List[FinetuneData]]]

    _event_model: Optional[Type[EventResult]] = EventResult
    _data_model: Optional[Type[FinetuneData]] = FinetuneData

    _request: Optional[FinetuneRequest] = None

    fine_tuned_model: Optional[str]
    hyperparams: Optional[Dict[str, Any]]
    organization_id: Optional[str]
    result_files: Optional[List[FileData]]
    training_files: Optional[List[FileData]]
    validation_files: Optional[List[FileData]]
    
    status: Optional[str]
    updated_at: Optional[datetime.datetime]

    @property
    def metadata_fields(self):
        return [
            'id',
            'object',
            'created_at',
            'fine_tuned_model',
            'hyperparams',
            'organization_id',
            'result_files',
            'training_files',
            'validation_files',
            'status',
            'updated_at',
            # 'events',
            # 'data',
        ]




import json
import pathlib
from enum import Enum
from pydantic import BaseSettings, validator

from typing import Optional, Dict, Any, Union

from async_openai.version import VERSION
from async_openai.types.classprops import lazyproperty


class ApiType(str, Enum):
    azure = "azure"
    openai = "openai"
    open_ai = "openai"
    azure_ad = "azure_ad"
    azuread = "azure_ad"

    def get_version(self, version: Optional[str] = None):
        if self.value in {"azure", "azure_ad", "azuread"}:
            return "2022-03-01-preview"
        return version
        

_should_reset_api: bool = False

class OpenAISettings(BaseSettings):
    api_key: str = None
    api_base: str = "https://api.openai.com/v1"
    organization: Optional[str] = None
    api_type: Optional[ApiType] = ApiType.open_ai
    api_version: Optional[str] = None
    api_key_path: Optional[pathlib.Path] = None

    verify_ssl_certs: Optional[bool] = True  # No effect. Certificates are always verified.
    proxy: Optional[str] = None
    app_info: Optional[Dict[str, str]] = None
    enable_telemetry: Optional[bool] = False  # Ignored; the telemetry feature was removed.
    ca_bundle_path: Optional[pathlib.Path] = None  # No longer used, feature was removed
    debug: Optional[bool] = False
    log_level: Optional[str] = None  # Set to either 'debug' or 'info', controls console logging

    timeout_secs: Optional[int] = 600
    max_retries: Optional[int] = 3


    @validator("api_type")
    def validate_api_type(cls, v):
        if v is None: return ApiType.open_ai
        return ApiType(v) if isinstance(v, str) else v


    @validator("api_version")
    def validate_api_version(cls, v, values):
        if 'api_type' in values and values['api_type'] in {"azure", "azure_ad", "azuread"}:
            return "2022-03-01-preview"
        return v
    
    
    @validator("api_key")
    def validate_api_key(cls, v, values):
        if v is None and values.get('api_key_path') is not None:
            return values['api_key_path'].read_text()
        return v

    def should_reset_api(
        self,
    ):
        global _should_reset_api
        if _should_reset_api:
            _should_reset_api = False
            return True
        return False

    def configure(
        self, 
        api_key: Optional[str] = None,
        api_base: Optional[str] = None,
        api_type: Optional[Union[str, ApiType]] = None,
        api_version: Optional[str] = None,
        organization: Optional[str] = None,
        proxies: Optional[Union[str, Dict]] = None,
        timeout_secs: Optional[int] = None,
        **kwargs
    ):
        """
        Allows Post-Init Configuration of the OpenAI Settings
        """
        if api_key is not None: self.api_key = api_key
        if api_base is not None: self.api_base = api_base
        if api_type is not None: 
            self.api_type = api_type if isinstance(api_type, ApiType) else ApiType(api_type)
        if api_version is not None: self.api_version = self.api_type.get_version(api_version)
        if organization is not None: self.organization = organization
        if proxies is not None: self.proxies = proxies
        if timeout_secs is not None: self.timeout_secs = timeout_secs
        
        for k, v in kwargs.items():
            if not hasattr(self, k):  continue
            if isinstance(getattr(self, k), pathlib.Path):
                setattr(self, k, pathlib.Path(v))
            else:
                setattr(self, k, v)

    
    @lazyproperty
    def user_agent(self) -> Dict[str, str]:
        import platform
        ua = f"OpenAI/v1 async_openai/{VERSION}"
        if self.app_info:
            t = ""
            if "name" in self.app_info:
                t += self.app_info["name"]
            if "version" in self.app_info:
                t += f"/{self.app_info['version']}"
            if "url" in self.app_info:
                t += f" ({self.app_info['url']})"
            ua += f" {t}"
        uname_without_node = " ".join(
            v for k, v in platform.uname()._asdict().items() if k != "node"
        )
        data = {
            "bindings_version": VERSION,
            "httplib": "httpx",
            "lang": "python",
            "lang_version": platform.python_version(),
            "platform": platform.platform(),
            "publisher": "growthengineai",
            "uname": uname_without_node,
        }
        if self.app_info:
            data["application"] = self.app_info
        return {"X-OpenAI-Client-User-Agent": json.dumps(data), "User-Agent": ua}

    def api_key_headers(self, api_key: Optional[str] = None, api_type: Optional[Union[ApiType, str]] = None) -> Dict[str, str]:
        if api_key is None: api_key = self.api_key
        if api_type is None: api_type = self.api_type

        api_type = api_type.value if isinstance(api_type, ApiType) else api_type
        #if api_type in {"azure", "azure_ad", "azuread"}:
        return {"Authorization": f"Bearer {api_key}"}
        # return {"api-key": api_key}

    @lazyproperty
    def base_headers(self) -> Dict[str, Any]:
        """
        Returns the Base Headers for the Client
        """
        headers = self.user_agent
        headers.update(self.api_key_headers())
        if self.organization:
            headers["OpenAI-Organization"] = self.organization
        if self.api_version is not None and self.api_type == 'open_ai':
            headers["OpenAI-Version"] = self.api_version
        if self.debug:
            headers["OpenAI-Debug"] = "true"
        return headers

    def contruct_headers(
        self, 
        api_key: Optional[str] = None, 
        api_version: Optional[str] = None,
        api_type: Optional[Union[ApiType, str]] = None,
        organization: Optional[str] = None,
    ) -> Dict[str, str]:
        headers = self.user_agent
        headers.update(self.api_key_headers(api_key=api_key, api_type=api_type))
        if organization is None: organization = self.organization
        if api_version is None: api_version = self.api_version
        if api_type is None: api_type = self.api_type
        if organization:
            headers["OpenAI-Organization"] = organization
        if api_version is not None and api_type.value == 'open_ai':
            headers["OpenAI-Version"] = api_version
        if self.debug:
            headers["OpenAI-Debug"] = "true"
        return headers


    class Config:
        env_prefix = "OPENAI_"


settings = OpenAISettings()





from __future__ import absolute_import

from async_openai.schemas.types.base import FilePurpose, FileData
from async_openai.schemas.types.base import BaseResult, EventResult, Permission, Usage, BaseEndpoint

from async_openai.schemas.types.completions import CompletionRequest, CompletionResult, CompletionChoice, CompletionModels
from async_openai.schemas.types.edits import EditRequest, EditResult, EditChoice
from async_openai.schemas.types.embeddings import EmbeddingRequest, EmbeddingResult, EmbeddingData
from async_openai.schemas.types.engines import EngineRequest, EngineResult, EngineData

from async_openai.schemas.types.files import FileRequest, FileResult, FileData
from async_openai.schemas.types.finetunes import FinetuneRequest, FinetuneResult, FinetuneData
from async_openai.schemas.types.images import ImageRequest, ImageResult, ImageData
from async_openai.schemas.types.models import ModelRequest, ModelResult, ModelData



from typing import Optional
from async_openai.schemas.base import BaseSchema
from async_openai.schemas.types.edits import *


class EditSchema(BaseSchema):

    def create(
        self,
        model: Optional[str],
        instruction: Optional[str],
        input: Optional[str] = "",
        n: Optional[int] = 1,
        temperature: Optional[float] = 1.0,
        top_p: Optional[float] = 1.0,
        user: Optional[str] = None,
        **kwargs,
    ) -> EditResult:
        """
        Create an Edit.
        """
        request = EditRequest(
            model = model,
            instruction = instruction,
            input = input,
            n = n,
            temperature = temperature,
            top_p = top_p,
            user = user,
        )
        response = self.send(
            **request.create_edit_endpoint.get_params(**kwargs)
        )
        result = EditResult(
            _raw_request = request,
            _raw_response = response,
        )
        result.parse_result()
        return result
    
    async def async_create(
        self,
        model: Optional[str],
        instruction: Optional[str],
        input: Optional[str] = "",
        n: Optional[int] = 1,
        temperature: Optional[float] = 1.0,
        top_p: Optional[float] = 1.0,
        user: Optional[str] = None,
        **kwargs,
    ) -> EditResult:
        """
        [Async] Create an Edit.
        """
        request = EditRequest(
            model = model,
            instruction = instruction,
            input = input,
            n = n,
            temperature = temperature,
            top_p = top_p,
            user = user,
        )
        response = await self.async_send(
            **await request.create_edit_endpoint.async_get_params(**kwargs)
        )
        result = EditResult(
            _raw_request = request,
            _raw_response = response,
        )
        result.parse_result()
        return result

    






# from .utils import merge_dicts
from async_openai.utils.configs import logger

class OpenAIError(Exception):
    def __init__(
        self,
        message=None,
        http_body=None,
        http_status=None,
        json_body=None,
        headers=None,
        code=None,
    ):
        super(OpenAIError, self).__init__(message)

        if http_body and hasattr(http_body, "decode"):
            try:
                http_body = http_body.decode("utf-8")
            except BaseException:
                http_body = (
                    "<Could not decode body as utf-8. "
                    "Please report to support@openai.com>"
                )

        self._message = message
        self.http_body = http_body
        self.http_status = http_status
        self.json_body = json_body
        self.headers = headers or {}
        self.code = code
        self.request_id = self.headers.get("request-id", None)
        self.error = self.construct_error_object()
        self.organization = self.headers.get("openai-organization", None)

    def __str__(self):
        msg = self._message or "<empty message>"
        if self.request_id is not None:
            return "Request {0}: {1}".format(self.request_id, msg)
        else:
            return msg

    # Returns the underlying `Exception` (base class) message, which is usually
    # the raw message returned by OpenAI's API. This was previously available
    # in python2 via `error.message`. Unlike `str(error)`, it omits "Request
    # req_..." from the beginning of the string.
    @property
    def user_message(self):
        return self._message

    def __repr__(self):
        return "%s(message=%r, http_status=%r, request_id=%r)" % (
            self.__class__.__name__,
            self._message,
            self.http_status,
            self.request_id,
        )

    def construct_error_object(self):
        if (
            self.json_body is None
            or "error" not in self.json_body
            or not isinstance(self.json_body["error"], dict)
        ):
            return None
        
        # values = merge_dicts({"message": None, "type": None}, values)


        # return openai.api_resources.error_object.ErrorObject.construct_from(
        #     self.json_body["error"]
        # )


class APIError(OpenAIError):
    pass


class TryAgain(OpenAIError):
    pass


class Timeout(OpenAIError):
    pass


class APIConnectionError(OpenAIError):
    def __init__(
        self,
        message,
        http_body=None,
        http_status=None,
        json_body=None,
        headers=None,
        code=None,
        should_retry=False,
    ):
        super(APIConnectionError, self).__init__(
            message, http_body, http_status, json_body, headers, code
        )
        self.should_retry = should_retry


class InvalidRequestError(OpenAIError):
    def __init__(
        self,
        message,
        param,
        code=None,
        http_body=None,
        http_status=None,
        json_body=None,
        headers=None,
    ):
        super(InvalidRequestError, self).__init__(
            message, http_body, http_status, json_body, headers, code
        )
        self.param = param

    def __repr__(self):
        return "%s(message=%r, param=%r, code=%r, http_status=%r, " "request_id=%r)" % (
            self.__class__.__name__,
            self._message,
            self.param,
            self.code,
            self.http_status,
            self.request_id,
        )

    def __reduce__(self):
        return type(self), (
            self._message,
            self.param,
            self.code,
            self.http_body,
            self.http_status,
            self.json_body,
            self.headers,
        )


class AuthenticationError(OpenAIError):
    pass


class PermissionError(OpenAIError):
    pass


class RateLimitError(OpenAIError):
    pass


class ServiceUnavailableError(OpenAIError):
    pass


class InvalidAPIType(OpenAIError):
    pass


class SignatureVerificationError(OpenAIError):
    def __init__(self, message, sig_header, http_body=None):
        super(SignatureVerificationError, self).__init__(message, http_body)
        self.sig_header = sig_header

    def __reduce__(self):
        return type(self), (
            self._message,
            self.sig_header,
            self.http_body,
        )


def handle_error_response(
    rbody, 
    rcode, 
    resp, 
    rheaders, 
    stream_error=False
):
    try:
        error_data = resp["error"]
    except (KeyError, TypeError) as e:
        raise APIError("Invalid response object from API: %r (HTTP response code " "was %d)" % (rbody, rcode), rbody, rcode, resp,) from e

    if "internal_message" in error_data:
        error_data["message"] += "\n\n" + error_data["internal_message"]

    logger(
        "OpenAI API error received",
        error_code=error_data.get("code"),
        error_type=error_data.get("type"),
        error_message=error_data.get("message"),
        error_param=error_data.get("param"),
        stream_error=stream_error,
        level = "error"
    )

    # Rate limits were previously coded as 400's with code 'rate_limit'
    if rcode == 429:
        return RateLimitError(
            error_data.get("message"), rbody, rcode, resp, rheaders
        )
    elif rcode in [400, 404, 415]:
        return InvalidRequestError(
            error_data.get("message"),
            error_data.get("param"),
            error_data.get("code"),
            rbody,
            rcode,
            resp,
            rheaders,
        )
    elif rcode == 401:
        return AuthenticationError(
            error_data.get("message"), rbody, rcode, resp, rheaders
        )
    elif rcode == 403:
        return PermissionError(
            error_data.get("message"), rbody, rcode, resp, rheaders
        )
    elif rcode == 409:
        return TryAgain(
            error_data.get("message"), rbody, rcode, resp, rheaders
        )
    elif stream_error:
        # TODO: we will soon attach status codes to stream errors
        parts = [error_data.get("message"), "(Error occurred while streaming.)"]
        message = " ".join([p for p in parts if p is not None])
        return APIError(message, rbody, rcode, resp, rheaders)
    else:
        return APIError(
            f"{error_data.get('message')} {rbody} {rcode} {resp} {rheaders}",
            rbody,
            rcode,
            resp,
            rheaders,
        )


# def interpret_response_line(
#     rbody, 
#     rcode, 
#     rheaders, 
#     stream: bool
# ) -> OpenAIResponse:
#     # HTTP 204 response code does not have any content in the body.
#     if rcode == 204:
#         return OpenAIResponse(None, rheaders)

#     if rcode == 503:
#         raise ServiceUnavailableError(
#             "The server is overloaded or not ready yet.",
#             rbody,
#             rcode,
#             headers=rheaders,
#         )
#     try:
#         if hasattr(rbody, "decode"):
#             rbody = rbody.decode("utf-8")
#         data = json.loads(rbody)
#     except (json.JSONDecodeError, UnicodeDecodeError):
#         raise APIError(
#             f"HTTP code {rcode} from API ({rbody})", rbody, rcode, headers=rheaders
#         )
#     resp = OpenAIResponse(data, rheaders)
#     # In the future, we might add a "status" parameter to errors
#     # to better handle the "error while streaming" case.
#     stream_error = stream and "error" in resp.data
#     if stream_error or not 200 <= rcode < 300:
#         raise handle_error_response(
#             rbody, rcode, resp.data, rheaders, stream_error=stream_error
#         )
#     return resp


from __future__ import absolute_import

from async_openai.schemas.base import BaseSchema
from async_openai.schemas.client import Client
from async_openai.schemas.completion import CompletionSchema
from async_openai.schemas.edits import EditSchema
from async_openai.schemas.embeddings import EmbeddingSchema




from enum import Enum
from pydantic import validator
from typing import List, Any, Optional, Union, Dict, Type
from async_openai.types import BaseModel, lazyproperty
from async_openai.schemas.types.base import BaseResult, Method, BaseEndpoint

from fileio import File, FileType

__all__ = [
    'ImageSize',
    'ImageFormat',
    'ImageRequest',
    'ImageData',
    'ImageResult',
]

class ImageSize(str, Enum):
    """
    Size of the image
    """

    small = "256x256"
    medium = "512x512"
    large = "1024x1024"

    @classmethod
    def from_str(cls, value: str) -> "ImageSize":
        """
        :param value: Size of the image
        :type value: str
        :return: ImageSize
        :rtype: ImageSize
        """
        if value == "256x256":
            return cls.small
        if value == "512x512":
            return cls.medium
        if value == "1024x1024":
            return cls.large
        raise ValueError(f"Cannot convert {value} to ImageSize")

class ImageFormat(str, Enum):
    """
    Format of the image
    """

    url = "url"
    b64 = "b64_json"
    b64_json = "b64_json"


class ImageRequest(BaseModel):
    prompt: Optional[str]
    mask: Optional[Union[str, FileType, Any]]
    image: Optional[Union[str, FileType, Any]]
    n: Optional[int] = 1
    size: Optional[Union[str, ImageSize]] = ImageSize.large
    response_format: Optional[Union[str, ImageFormat]] = ImageFormat.url
    user: Optional[str] = None

    @validator("size")
    def validate_size(cls, value):
        return ImageSize.from_str(value) if isinstance(value, str) else value
    
    @validator("response_format")
    def validate_response_format(cls, value):
        if isinstance(value, str):
            value = ImageFormat(value)
        return value
    

    def get_file_request_params(self, **kwargs) -> Dict:
        """
        Transforms the data to the req params
        """
        files = [(k, (None, v)) for k, v in self.dict(exclude_none=True, exclude={'mask', 'image'}).items()]
        if self.mask:
            mask = File(self.mask)
            files.append(("mask", ("mask", mask.read_bytes(), "application/octet-stream")))
        if self.image:
            image = File(self.image)
            files.append(("image", ("image", image.read_bytes(), "application/octet-stream")))
        
        return files

    
    async def async_get_file_request_params(self, **kwargs) -> Dict:
        """
        Transforms the data to the req params
        """
        files = [(k, (None, v)) for k, v in self.dict(exclude_none=True, exclude={'mask', 'image'}).items()]
        if self.mask:
            mask = File(self.mask)
            files.append(("mask", ("mask", await mask.async_read_bytes(), "application/octet-stream")))
        if self.image:
            image = File(self.image)
            files.append(("image", ("image", await image.async_read_bytes(), "application/octet-stream")))
        
        return files

    
    @lazyproperty
    def create_image_endpoint(self) -> BaseEndpoint:
        return BaseEndpoint(
            method = Method.POST,
            url = '/images/generations',
            files = self.async_get_file_request_params,
            sync_files = self.get_file_request_params
        )
    
    @lazyproperty
    def create_image_edit_endpoint(self) -> BaseEndpoint:
        return BaseEndpoint(
            method = Method.POST,
            url = '/images/edits',
            files = self.async_get_file_request_params,
            sync_files = self.get_file_request_params
        )
    
    @lazyproperty
    def create_image_variation_endpoint(self) -> BaseEndpoint:
        return BaseEndpoint(
            method = Method.POST,
            url = '/images/variations',
            files = self.async_get_file_request_params,
            sync_files = self.get_file_request_params
        )
    

class ImageData(BaseModel):
    url: Optional[str] = None
    data: Optional[bytes] = None


class ImageResult(BaseResult):
    data: Optional[Union[ImageData, List[ImageData]]]
    _data_model: Optional[Type[ImageData]] = ImageData
    _request: Optional[ImageRequest] = None

    @property
    def metadata_fields(self):
        return [
            'created',
            'object',
            # 'data',
        ]




from typing import Dict, Union, Optional

def parse_stream(body: Union[str, bytes]):
    for line in body:
        if line:
            if line == b"data: [DONE]":
                # return here will cause GeneratorExit exception in urllib3
                # and it will close http connection with TCP Reset
                continue
            if hasattr(line, "decode"):
                line = line.decode("utf-8")
            if line.startswith("data: "):
                line = line[len("data: ") :]
            yield line


def build_proxies(
    proxy: Optional[Union[str, Dict]] = None,
):
    if proxy is None:
        return None
    elif isinstance(proxy, str):
        return {"http://": proxy, "https://": proxy}
    elif isinstance(proxy, dict):
        return proxy.copy()
    else:
        raise TypeError("Proxy must be a string or a dictionary")





import inspect
from re import sub


def is_coro_func(obj, func_name: str = None):
    """
    This is probably in the library elsewhere but returns bool
    based on if the function is a coro
    """
    try:
        if inspect.iscoroutinefunction(obj): return True
        if inspect.isawaitable(obj): return True
        if func_name and hasattr(obj, func_name) and inspect.iscoroutinefunction(getattr(obj, func_name)):
            return True
        return bool(hasattr(obj, '__call__') and inspect.iscoroutinefunction(obj.__call__))
    except Exception:
        return False

def to_camel_case(text: str):
    """Convert a snake str to camel case."""
    components = text.split("_")
    # We capitalize the first letter of each component except the first one
    # with the 'title' method and join them together.
    return components[0] + "".join(x.title() for x in components[1:])

def to_snake_case(text: str):
    """
    multiMaster -> multi_master
    """
    return '_'.join(
        sub('([A-Z][a-z]+)', r' \1',
        sub('([A-Z]+)', r' \1',
        text.replace('-', ' '))).split()).lower()

def to_snake_case_args(text: str):
    """
    multiMaster -> multi-master
    """
    return '-'.join(
        sub('([A-Z][a-z]+)', r' \1',
        sub('([A-Z]+)', r' \1',
        text.replace('-', ' '))).split()).lower()



from typing import Dict, Union, Optional
from async_openai.schemas.types.exceptions import APIError, TryAgain, RateLimitError


def fatal_exception(exc):
    if isinstance(exc, (APIError, TryAgain)):
        # retry on server errors and client errors
        # with 429 status code (rate limited),
        # don't retry on other client errors
        return (400 <= exc.exc.status_code < 500) and exc.exc.status_code != 429
    else:
        # retry on all other errors (eg. network)
        return False


def build_proxies(
    proxy: Optional[Union[str, Dict]] = None,
):
    if proxy is None:
        return None
    elif isinstance(proxy, str):
        return {"http://": proxy, "https://": proxy}
    elif isinstance(proxy, dict):
        return proxy.copy()
    else:
        raise TypeError("Proxy must be a string or a dictionary")





from pydantic import Field
from pydantic import BaseModel as _BaseModel
from typing import Any

from async_openai.types.formatting import to_snake_case

class BaseModel(_BaseModel):
    class Config:
        extra = 'allow'
        arbitrary_types_allowed = True
        alias_generator = to_snake_case

    def get(self, name, default: Any = None):
        return getattr(self, name, default)


class Schema(BaseModel):

    class Config:
        extra = 'allow'
        arbitrary_types_allowed = True
        

    def get(self, name, default: Any = None):
        return getattr(self, name, default)


__all__ = [
    'Field',
    'BaseModel',
    'Schema',
]



# from .utils import merge_dicts
import httpx
import json
from typing import Any, Optional, Union, Dict
from async_openai.types import BaseModel, lazyproperty

# We'll use this to construct our exception
class ExceptionModel(BaseModel):
    response: httpx.Response
    message: Optional[str]
    data: Optional[Union[Dict, Any]]
    stream: Optional[bool] = False
    should_retry: Optional[bool] = False

    @lazyproperty
    def headers(self):
        return self.response.headers

    @lazyproperty
    def response_data(self):
        return self.data or self.response.json()
    
    @lazyproperty
    def http_body(self):
        body = self.data if self.stream else self.response.content
        try:
            if hasattr(body, "decode"):
                body = body.decode("utf-8")
            return body
        except (json.JSONDecodeError, UnicodeDecodeError) as e:
            raise ValueError(
                f"HTTP code {self.status_code} from API ({body})"
            ) from e
    
    @lazyproperty
    def response_json(self):
        try:
            return json.loads(self.http_body)
        except json.JSONDecodeError:
            return {}

    @lazyproperty
    def response_text(self):
        return self.response.text
    
    @lazyproperty
    def status_code(self):
        return self.response.status_code
    
    @lazyproperty
    def error_data(self) -> Dict:
        return self.response_json.get("error", {})
    
    @lazyproperty
    def request_id(self) -> str:
        return self.headers.get("request-id", None)
    
    @lazyproperty
    def organization(self) -> str:
        return self.headers.get("openai-organization", None)
    
    @lazyproperty
    def error_message(self) -> str:
        msg = self.message or ""
        msg += " " + self.error_data.get("message", "")
        if self.error_data.get("internal_message"):
            msg += "\n\n" + self.error_data["internal_message"]
        return msg



class OpenAIError(Exception):
    def __init__(
        self,
        exc: ExceptionModel,
        **kwargs
    ):
        super(OpenAIError, self).__init__(exc.error_message)
        self.exc = exc
    
    def __str__(self):
        msg = self.exc.error_message or "<empty message>"
        if self.exc.request_id is not None:
            return "Request {0}: {1}".format(self.exc.request_id, msg)
        else:
            return msg
    
    # Returns the underlying `Exception` (base class) message, which is usually
    # the raw message returned by OpenAI's API. This was previously available
    # in python2 via `error.message`. Unlike `str(error)`, it omits "Request
    # req_..." from the beginning of the string.
    @property
    def user_message(self):
        return self.exc.error_message
    
    def __repr__(self):
        return "%s(message=%r, http_status=%r, request_id=%r)" % (
            self.__class__.__name__,
            self.exc.error_message,
            self.exc.status_code,
            self.exc.request_id,
        )


class APIError(OpenAIError):
    pass


class TryAgain(OpenAIError):
    pass


class Timeout(OpenAIError):
    pass


class APIConnectionError(OpenAIError):
    pass


class InvalidRequestError(OpenAIError):
    pass


class AuthenticationError(OpenAIError):
    pass


class PermissionError(OpenAIError):
    pass


class RateLimitError(OpenAIError):
    pass


class ServiceUnavailableError(OpenAIError):
    pass


class InvalidAPIType(OpenAIError):
    pass


def error_handler(
    response: httpx.Response,
    data: Optional[Any] = None, # Line specific for streaming responses
    stream: bool = False,
    should_retry: Optional[bool] = False,
    **kwargs
):
    exc = ExceptionModel(
        response = response,
        data = data,
        stream = stream,
        should_retry = should_retry,
        **kwargs
    )
    if response.status_code == 503:
        exc.message = "The server is overloaded or not ready yet."
        raise ServiceUnavailableError(
            exc = exc
        )
    if response.status_code == 429:
        raise RateLimitError(
            exc = exc
        )
    if response.status_code in [400, 404, 415]:
        return InvalidRequestError(
            exc = exc
        )
    if response.status_code == 401:
        return AuthenticationError(
            exc = exc
        )
    if response.status_code == 403:
        return PermissionError(
            exc = exc
        )
    if response.status_code == 409:
        return TryAgain(
            exc = exc
        )
    
    if stream: exc.message = "(Error occurred while streaming.)"
    raise APIError(exc = exc)






from typing import List, Any, Optional, Union, Dict, Type
from async_openai.types import BaseModel, lazyproperty
from async_openai.schemas.types.base import BaseResult, Method, BaseEndpoint

__all__ = [
    'EditChoice',
    'EditRequest',
    'EditResult',
]

class EditChoice(BaseModel):
    text: str
    index: int
    logprobs: Optional[Any]
    finish_reason: Optional[str]

class EditRequest(BaseModel):
    model: Optional[str]
    instruction: Optional[str]
    input: Optional[str] = ""
    n: Optional[int] = 1
    temperature: Optional[float] = 1.0
    top_p: Optional[float] = 1.0
    user: Optional[str] = None

    @property
    def create_edit_endpoint(self) -> BaseEndpoint:
        return BaseEndpoint(
            method = Method.POST,
            url ='/edits',
            data = self.dict(
                exclude_none = True
            )
        )


class EditResult(BaseResult):
    choices: Optional[List[EditChoice]]
    _choice_model: Optional[Type[EditChoice]] = EditChoice
    _request: Optional[EditRequest] = None








import backoff
from async_openai.types import BaseModel

from async_openai.schemas.client import Client
from async_openai.schemas.utils import fatal_exception
from async_openai.utils import settings
from async_openai.schemas.types.base import *

__all__  = [
    'BaseSchema',
]

class BaseSchema(BaseModel):
    """
    Base class for all schemas.
    """

    @backoff.on_exception(backoff.expo, Exception, max_tries = settings.max_retries + 1, giveup = fatal_exception)
    def send(
        self,
        **kwargs,
    ):
        """
        Send the request.
        """
        headers = Client.get_headers(**kwargs)
        return Client.request(
            headers = headers,
            **kwargs
        )

    @backoff.on_exception(backoff.expo, Exception, max_tries = settings.max_retries + 1, giveup = fatal_exception)
    async def async_send(
        self,
        **kwargs,
    ):
        """
        [Async] Send the request.
        """
        headers = Client.get_headers(**kwargs)
        return await Client.async_request(
            headers = headers,
            **kwargs
        )
        



from typing import Optional, Union, List, Any
from async_openai.schemas.base import BaseSchema
from async_openai.schemas.types.embeddings import *


class EmbeddingSchema(BaseSchema):

    def create(
        self,
        model: Optional[str],
        input: Optional[Union[List[Any], Any]],
        user: Optional[str] = None,
        **kwargs,
    ) -> EmbeddingResult:
        """
        Create an Embedding.
        """
        request = EmbeddingRequest(
            model = model,
            input = input,
            user = user,
        )
        response = self.send(
            **request.create_embeddings_endpoint.get_params(**kwargs)
        )
        result = EmbeddingResult(
            _raw_request = request,
            _raw_response = response,
        )
        result.parse_result()
        return result
    
    async def async_create(
        self,
        model: Optional[str],
        input: Optional[Union[List[Any], Any]],
        user: Optional[str] = None,
        **kwargs,
    ) -> EmbeddingResult:
        """
        [Async] Create an Embedding.
        """
        request = EmbeddingRequest(
            model = model,
            input = input,
            user = user,
        )
        response = await self.async_send(
            **await request.create_embeddings_endpoint.async_get_params(**kwargs)
        )
        result = EmbeddingResult(
            _raw_request = request,
            _raw_response = response,
        )
        result.parse_result()
        return result

    





import json
import httpx

from enum import Enum
from pydantic import validator
from typing import Optional, Union, List, Dict, Any, Iterator, Tuple, TypeVar, Callable, cast, overload

from async_openai.types import BaseModel
from async_openai.utils.configs import settings, ApiType, logger
from async_openai.client.utils import build_proxies, parse_stream
from async_openai.client.exceptions import *

class Method(str, Enum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    PATCH = "PATCH"
    HEAD = "HEAD"


class OpenAIResponse(BaseModel):
    data: Any
    headers: Optional[Dict[str, Any]] = {}

    @property
    def request_id(self) -> Optional[str]:
        return self.headers.get("request-id")

    @property
    def organization(self) -> Optional[str]:
        return self.headers.get("OpenAI-Organization")

    @property
    def response_ms(self) -> Optional[int]:
        h = self.headers.get("Openai-Processing-Ms")
        return None if h is None else round(float(h))
    


class APIRequestor:
    def __init__(
        self, 
        api_key: Optional[str] =  settings.api_key,
        api_base: Optional[str] = settings.api_base,
        api_type: Optional[Union[str, ApiType]] = settings.api_type,
        api_version: Optional[str] = None,
        organization: Optional[str] = settings.organization,
        proxies: Optional[Union[str, Dict]] = settings.proxy,
        timeout_secs: Optional[int] = settings.timeout_secs,
        **kwargs
    ):
        self.api_key = api_key or settings.api_key
        self.api_base = api_base or settings.api_base
        self.api_type = api_type if isinstance(api_type, ApiType) else ApiType(api_type)
        self.api_version = self.api_type.get_version(api_version) if api_version else settings.api_version
        self.organization = organization or settings.organization
        self.proxies = build_proxies(proxies or settings.proxy)
        self.timeout_secs = timeout_secs or settings.timeout_secs

        self.base_headers = settings.contruct_headers(
            api_key = self.api_key,
            api_type = self.api_type,
            api_version = self.api_version,
            organization = self.organization,
        )
        self._aclient = httpx.AsyncClient(
            base_url = self.api_base,
            proxies = self.proxies,
            timeout = self.timeout_secs,
        )
        self._client = httpx.Client(
            base_url = self.api_base,
            proxies = self.proxies,
            timeout = self.timeout_secs,
        )
    
    def create_headers(
        self,
        supplied_headers: Optional[Dict[str, str]] = None,
        request_id: Optional[str] = None,
        **kwargs,
    ) -> Dict[str, str]:
        """
        Create the headers for the request
        """
        headers = self.base_headers.copy()
        if supplied_headers: headers.update(supplied_headers)
        if request_id: headers["X-Request-Id"] = request_id
        return headers

    def create_request_params(
        self,
        method: Method,
        url: str,
        *,
        params: Optional[Dict[str, Any]] = None,
        supplied_headers: Optional[Dict[str, str]] = None,
        files: Optional[Any] = None,
        request_id: Optional[str] = None,
        request_timeout: Optional[Union[float, Tuple[float, float]]] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Unified method for constructing the request parameters
        """
        method = Method(method) if isinstance(method, str) else method
        headers = self.create_headers(supplied_headers = supplied_headers, request_id = request_id)
        data = None

        if method in {method.GET, method.DELETE}:
            if params: params = {k: v for k, v in params.items() if v is not None}
        elif method in {method.POST, method.PUT}:
            if params and files: raise ValueError("At most one of params and files may be specified.")
            if params:
                data = json.dumps(params).encode()
                headers["Content-Type"] = "application/json"
                params = None
        else:
            raise APIConnectionError(
                "Unrecognized HTTP method %r. This may indicate a bug in the "
                "OpenAI bindings. Please contact support@openai.com for "
                "assistance." % (method.value,)
            )
        
        return {
            "method": method.value,
            "url": url,
            "params": params,
            "headers": headers,
            "data": data,
            "files": files,
            "timeout": request_timeout,
        }
    
    def _interpret_response_line(
        self, rbody, rcode, rheaders, stream: bool
    ) -> OpenAIResponse:
        # HTTP 204 response code does not have any content in the body.
        if rcode == 204:
            return OpenAIResponse(None, rheaders)

        if rcode == 503:
            raise ServiceUnavailableError(
                "The server is overloaded or not ready yet.",
                rbody,
                rcode,
                headers=rheaders,
            )
        try:
            if hasattr(rbody, "decode"):
                rbody = rbody.decode("utf-8")
            data = json.loads(rbody)
        except (json.JSONDecodeError, UnicodeDecodeError) as e:
            raise APIError(f"HTTP code {rcode} from API ({rbody})", rbody, rcode, headers=rheaders) from e

        resp = OpenAIResponse(data, rheaders)
        # In the future, we might add a "status" parameter to errors
        # to better handle the "error while streaming" case.
        stream_error = stream and "error" in resp.data
        if stream_error or not 200 <= rcode < 300:
            raise handle_error_response(
                rbody, rcode, resp.data, rheaders, stream_error=stream_error
            )
        return resp

    def _interpret_response(
        self, result: httpx.Response, stream: bool
    ) -> Tuple[Union[OpenAIResponse, Iterator[OpenAIResponse]], bool]:
        """Returns the response(s) and a bool indicating whether it is a stream."""
        if stream and "text/event-stream" in result.headers.get("Content-Type", ""):
            return (
                self._interpret_response_line(
                    line, result.status_code, result.headers, stream=True
                )
                for line in parse_stream(result.iter_lines())
            ), True
        else:
            return (
                self._interpret_response_line(
                    result.content, result.status_code, result.headers, stream=False
                ),
                False,
            )


    def request_raw(
        self,
        method: Method,
        url: str,
        *,
        params: Optional[Dict[str, Any]] = None,
        supplied_headers: Optional[Dict[str, str]] = None,
        files: Optional[Any] = None,
        stream: Optional[bool] = False,
        request_id: Optional[str] = None,
        request_timeout: Optional[Union[float, Tuple[float, float]]] = None,
        **kwargs
    ) -> httpx.Response:
        """
        The sync raw request method.
        """
        request_params = self.create_request_params(
            method = method,
            url = url,
            params = params,
            supplied_headers = supplied_headers,
            files = files,
            request_id = request_id,
            request_timeout = request_timeout,
        )
        req_method = self._client.stream if stream else self._client.request
        try:
            return req_method(**request_params)
        except httpx.TimeoutException as e:
            raise APIConnectionError("Request timed out.") from e
        except httpx.HTTPError as e:
            raise APIConnectionError("Unexpected error communicating with OpenAI.") from e


    def request(
        self,
        method: Method,
        url: str,
        *,
        params: Optional[Dict[str, Any]] = None,
        supplied_headers: Optional[Dict[str, str]] = None,
        files: Optional[Any] = None,
        stream: Optional[bool] = False,
        request_id: Optional[str] = None,
        request_timeout: Optional[Union[float, Tuple[float, float]]] = None,
        **kwargs
    ) ->  Tuple[Union[OpenAIResponse, Iterator[OpenAIResponse]], bool, str]:
        """
        The sync request method.
        """
        result = self.request_raw(
            method = method,
            url = url,
            params = params,
            supplied_headers = supplied_headers,
            files = files,
            stream = stream,
            request_id = request_id,
            request_timeout = request_timeout,
        )
        resp, got_stream = self._interpret_response(result, stream)
        return resp, got_stream, self.api_key

    """
    Async Methods
    """

    async def async_request_raw(
        self,
        method: Method,
        url: str,
        *,
        params: Optional[Dict[str, Any]] = None,
        supplied_headers: Optional[Dict[str, str]] = None,
        files: Optional[Any] = None,
        stream: Optional[bool] = False,
        request_id: Optional[str] = None,
        request_timeout: Optional[Union[float, Tuple[float, float]]] = None,
        **kwargs
    ) -> httpx.Response:
        """
        The sync raw request method.
        """
        request_params = self.create_request_params(
            method = method,
            url = url,
            params = params,
            supplied_headers = supplied_headers,
            files = files,
            request_id = request_id,
            request_timeout = request_timeout,
        )
        req_method = self._aclient.stream if stream else self._aclient.request
        try:
            return req_method(**request_params)
        except httpx.TimeoutException as e:
            raise APIConnectionError("Request timed out.") from e
        except httpx.HTTPError as e:
            raise APIConnectionError("Unexpected error communicating with OpenAI.") from e

    async def async_request(
        self,
        method: Method,
        url: str,
        *,
        params: Optional[Dict[str, Any]] = None,
        supplied_headers: Optional[Dict[str, str]] = None,
        files: Optional[Any] = None,
        stream: Optional[bool] = False,
        request_id: Optional[str] = None,
        request_timeout: Optional[Union[float, Tuple[float, float]]] = None,
        **kwargs
    ) ->  Tuple[Union[OpenAIResponse, Iterator[OpenAIResponse]], bool, str]:
        """
        The sync request method.
        """
        result = await self.async_request_raw(
            method = method,
            url = url,
            params = params,
            supplied_headers = supplied_headers,
            files = files,
            stream = stream,
            request_id = request_id,
            request_timeout = request_timeout,
        )
        resp, got_stream = self._interpret_response(result, stream)
        return resp, got_stream, self.api_key


class Client:
    api: APIRequestor = None
    
    @classmethod
    def configure(
        cls,
        api_key: Optional[str] = None,
        api_base: Optional[str] = None,
        api_type: Optional[Union[str, ApiType]] = None,
        api_version: Optional[str] = None,
        organization: Optional[str] = None,
        proxies: Optional[Union[str, Dict]] = None,
        timeout_secs: Optional[int] = None,
        **kwargs
    ):
        _reset = api_base and api_base != settings.api_base
        settings.configure(
            api_key = api_key,
            api_base = api_base,
            api_type = api_type,
            api_version = api_version,
            organization = organization,
            proxies = proxies,
            timeout_secs = timeout_secs,
            **kwargs
        )
        cls.init_api(_reset = _reset)

    @classmethod
    def init_api(
        cls,
        _reset: bool = False,
        **kwargs
    ):
        if cls.api is None or _reset: cls.api = APIRequestor(**kwargs)
    
    @classmethod
    def request(
        cls,
        method: Method,
        url: str,
        *args,
        params: Optional[Dict[str, Any]] = None,
        supplied_headers: Optional[Dict[str, str]] = None,
        files: Optional[Any] = None,
        stream: Optional[bool] = False,
        request_id: Optional[str] = None,
        request_timeout: Optional[Union[float, Tuple[float, float]]] = None,
        **kwargs
    ) ->  Tuple[Union[OpenAIResponse, Iterator[OpenAIResponse]], bool, str]:
        """
        The sync request method.
        """
        cls.init_api()
        return cls.api.request(
            *args,
            method = method,
            url = url,
            params = params,
            supplied_headers = supplied_headers,
            files = files,
            stream = stream,
            request_id = request_id,
            request_timeout = request_timeout,
            **kwargs
        )

    @classmethod
    async def async_request(
        cls,
        method: Method,
        url: str,
        *args,
        params: Optional[Dict[str, Any]] = None,
        supplied_headers: Optional[Dict[str, str]] = None,
        files: Optional[Any] = None,
        stream: Optional[bool] = False,
        request_id: Optional[str] = None,
        request_timeout: Optional[Union[float, Tuple[float, float]]] = None,
        **kwargs
    ) ->  Tuple[Union[OpenAIResponse, Iterator[OpenAIResponse]], bool, str]:
        """
        The async request method.
        """
        cls.init_api()
        return await cls.api.async_request(
            *args,
            method = method,
            url = url,
            params = params,
            supplied_headers = supplied_headers,
            files = files,
            stream = stream,
            request_id = request_id,
            request_timeout = request_timeout,
            **kwargs
        )





from typing import Optional, Any, Union, List, Dict
from async_openai.schemas.base import BaseSchema
from async_openai.schemas.types.completions import *


class CompletionSchema(BaseSchema):

    def create(
        self,
        model: Optional[Union[str, Any]] = CompletionModels.curie,
        prompt: Optional[str] = '<|endoftext|>' ,
        suffix: Optional[str] = None,
        max_tokens: Optional[int] = 16,
        temperature: Optional[float] = 1.0,
        top_p: Optional[float] = 1.0,
        n: Optional[int] = 1,
        stream: Optional[bool] = False,
        logprobs: Optional[int] = None,
        echo: Optional[bool] = False,
        stop: Optional[Union[str, List[str]]] = None,
        presence_penalty: Optional[float] = 0.0,
        frequency_penalty: Optional[float] = 0.0,
        best_of: Optional[int] = 1,
        logit_bias: Optional[Dict[str, float]] = None,
        user: Optional[str] = None,
        **kwargs,
    ) -> CompletionResult:
        """
        Create a completion.
        """
        request = CompletionRequest(
            model = model,
            prompt = prompt,
            suffix = suffix,
            max_tokens = max_tokens,
            temperature = temperature,
            top_p = top_p,
            n = n,
            stream = stream,
            logprobs = logprobs,
            echo = echo,
            stop = stop,
            presence_penalty = presence_penalty,
            frequency_penalty = frequency_penalty,
            best_of = best_of,
            logit_bias = logit_bias,
            user = user,
        )
        response = self.send(
            **request.create_endpoint.get_params(**kwargs)
        )
        result = CompletionResult(
            _raw_request = request,
            _raw_response = response,
        )
        result.parse_result()
        return result
    
    async def async_create(
        self,
        model: Optional[Union[str, Any]] = CompletionModels.curie,
        prompt: Optional[str] = '<|endoftext|>' ,
        suffix: Optional[str] = None,
        max_tokens: Optional[int] = 16,
        temperature: Optional[float] = 1.0,
        top_p: Optional[float] = 1.0,
        n: Optional[int] = 1,
        stream: Optional[bool] = False,
        logprobs: Optional[int] = None,
        echo: Optional[bool] = False,
        stop: Optional[Union[str, List[str]]] = None,
        presence_penalty: Optional[float] = 0.0,
        frequency_penalty: Optional[float] = 0.0,
        best_of: Optional[int] = 1,
        logit_bias: Optional[Dict[str, float]] = None,
        user: Optional[str] = None,
        **kwargs,
    ) -> CompletionResult:
        """
        [Async] Create a completion.
        """
        request = CompletionRequest(
            model = model,
            prompt = prompt,
            suffix = suffix,
            max_tokens = max_tokens,
            temperature = temperature,
            top_p = top_p,
            n = n,
            stream = stream,
            logprobs = logprobs,
            echo = echo,
            stop = stop,
            presence_penalty = presence_penalty,
            frequency_penalty = frequency_penalty,
            best_of = best_of,
            logit_bias = logit_bias,
            user = user,
        )
        response = await self.async_send(
            **await request.create_endpoint.async_get_params(**kwargs)
        )
        result = CompletionResult(
            _raw_request = request,
            _raw_response = response,
        )
        result.parse_result()
        return result

    





from typing import List, Any, Optional, Union, Dict, Type
from async_openai.types import BaseModel, lazyproperty
from async_openai.schemas.types.base import BaseResult, Method, BaseEndpoint

__all__ = [
    'EmbeddingData',
    'EmbeddingRequest',
    'EmbeddingResult',
]


class EmbeddingData(BaseModel):
    object: Optional[str] = 'embedding'
    embedding: Optional[List[float]] = []
    index: Optional[int] = 0

class EmbeddingRequest(BaseModel):

    model: Optional[str]
    input: Optional[Union[List[Any], Any]]
    user: Optional[str] = None

    @property
    def create_embeddings_endpoint(self) -> BaseEndpoint:
        return BaseEndpoint(
            method = Method.POST,
            url = '/embeddings',
            data = self.dict(
                exclude_none = True
            )
        )


class EmbeddingResult(BaseResult):
    data: Optional[Union[EmbeddingData, List[EmbeddingData]]]
    _data_model: Optional[Type[EmbeddingData]] = EmbeddingData
    _request: Optional[EmbeddingRequest] = None

    @property
    def metadata_fields(self):
        return [
            'object',
            # 'data',
            'usage',
        ]





from enum import Enum
from pydantic import validator
from typing import List, Any, Optional, Union, Dict, Type
from async_openai.types import BaseModel, lazyproperty
from async_openai.schemas.types.base import BaseResult, BaseEndpoint, Method

__all__ = [
    'CompletionChoice',
    'CompletionModels',
    'CompletionRequest',
    'CompletionResult',
]

class CompletionChoice(BaseModel):
    text: str
    index: int
    logprobs: Optional[Any]
    finish_reason: Optional[str]

class CompletionModels(str, Enum):
    """
    Just the base models available
    """
    davinci = "text-davinci-003"
    curie = "text-curie-001"
    babbage = "text-babbage-001"
    ada = "text-ada-001"

class CompletionRequest(BaseModel):
    model: Optional[Union[str, Any]] = CompletionModels.curie
    prompt: Optional[str] = '<|endoftext|>'
    suffix: Optional[str] = None
    max_tokens: Optional[int] = 16
    temperature: Optional[float] = 1.0
    top_p: Optional[float] = 1.0
    n: Optional[int] = 1
    stream: Optional[bool] = False
    logprobs: Optional[int] = None
    echo: Optional[bool] = False
    stop: Optional[Union[str, List[str]]] = None
    presence_penalty: Optional[float] = 0.0
    frequency_penalty: Optional[float] = 0.0
    best_of: Optional[int] = 1
    logit_bias: Optional[Dict[str, float]] = None
    user: Optional[str] = None

    @validator('max_tokens')
    def validate_max_tokens(cls, v: int) -> int:
        """
        Max tokens is 4096
        https://beta.openai.com/docs/api-reference/completions/create#completions/create-max-tokens
        """
        return None if v is None else max(0, min(v, 4096))
    
    @validator('temperature')
    def validate_temperature(cls, v: float) -> float:
        """
        Min Temperature is 0.0
        https://beta.openai.com/docs/api-reference/completions/create#completions/create-temperature
        """
        return None if v is None else max(0.0, v)
    
    @validator('top_p')
    def validate_top_p(cls, v: float) -> float:
        """
        Min Top Probability is 0.0
        https://beta.openai.com/docs/api-reference/completions/create#completions/create-top-probability
        """
        return None if v is None else max(0.0, v)

    @validator('logprobs')
    def validate_logprobs(cls, v: int) -> int:
        """
        Max logprobs is 5
        https://beta.openai.com/docs/api-reference/completions/create#completions/create-logprobs
        """
        return None if v is None else max(0, min(v, 5))
    
    @validator('presence_penalty')
    def validate_presence_penalty(cls, v: float) -> float:
        """
        Min Presence Penalty is -2.0, Max is 2.0
        https://beta.openai.com/docs/api-reference/completions/create#completions/create-presence-penalty
        """
        return None if v is None else max(0.0, min(v, 2.0))
    
    @validator('frequency_penalty')
    def validate_frequency_penalty(cls, v: float) -> float:
        """
        Min Frequency Penalty is -2.0, Max is 2.0
        https://beta.openai.com/docs/api-reference/completions/create#completions/create-frequency-penalty
        """
        return None if v is None else max(0.0, min(v, 2.0))
    
    @property
    def create_endpoint(self) -> BaseEndpoint:
        return BaseEndpoint(
            method = Method.POST,
            url = '/completions',
            data = self.dict(exclude_none = True)
        )


class CompletionResult(BaseResult):
    choices: Optional[List[CompletionChoice]]

    _request: Optional[CompletionRequest] = None
    _choice_model: Optional[Type[CompletionChoice]] = CompletionChoice

    @property
    def completion_text(self) -> str:
        return ''.join(choice.text for choice in self.choices) if self.choices else ''














import httpx

from typing import Optional, Union, List, Dict

from async_openai.utils.configs import settings, ApiType, logger
from async_openai.schemas.utils import build_proxies


class APIRequestor:
    def __init__(
        self, 
        api_key: Optional[str] =  settings.api_key,
        api_base: Optional[str] = settings.api_base,
        api_type: Optional[Union[str, ApiType]] = settings.api_type,
        api_version: Optional[str] = None,
        organization: Optional[str] = settings.organization,
        proxies: Optional[Union[str, Dict]] = settings.proxy,
        timeout_secs: Optional[int] = settings.timeout_secs,
        **kwargs
    ):
        self.api_key = api_key or settings.api_key
        self.api_base = api_base or settings.api_base
        self.api_type = api_type if isinstance(api_type, ApiType) else ApiType(api_type)
        self.api_version = self.api_type.get_version(api_version) if api_version else settings.api_version
        self.organization = organization or settings.organization
        self.proxies = build_proxies(proxies or settings.proxy)
        self.timeout_secs = timeout_secs or settings.timeout_secs

        self.base_headers = settings.contruct_headers(
            api_key = self.api_key,
            api_type = self.api_type,
            api_version = self.api_version,
            organization = self.organization,
        )
        print(self.base_headers)
        self._aclient = httpx.AsyncClient(
            base_url = self.api_base,
            proxies = self.proxies,
            timeout = self.timeout_secs,
        )
        self._client = httpx.Client(
            base_url = self.api_base,
            proxies = self.proxies,
            timeout = self.timeout_secs,
        )
    
    def create_headers(
        self,
        supplied_headers: Optional[Dict[str, str]] = None,
        request_id: Optional[str] = None,
        **kwargs,
    ) -> Dict[str, str]:
        """
        Create the headers for the request
        """
        headers = self.base_headers.copy()
        if supplied_headers: headers.update(supplied_headers)
        if request_id: headers["X-Request-Id"] = request_id
        return headers
    


class Client:
    api: APIRequestor = None
    
    @classmethod
    def configure(
        cls,
        api_key: Optional[str] = None,
        api_base: Optional[str] = None,
        api_type: Optional[Union[str, ApiType]] = None,
        api_version: Optional[str] = None,
        organization: Optional[str] = None,
        proxies: Optional[Union[str, Dict]] = None,
        timeout_secs: Optional[int] = None,
        **kwargs
    ):
        _reset = api_base and api_base != settings.api_base
        settings.configure(
            api_key = api_key,
            api_base = api_base,
            api_type = api_type,
            api_version = api_version,
            organization = organization,
            proxies = proxies,
            timeout_secs = timeout_secs,
            **kwargs
        )
        cls.init_api(_reset = _reset)

    @classmethod
    def init_api(
        cls,
        _reset: bool = False,
        **kwargs
    ):
        if (cls.api is None or _reset or settings.should_reset_api()): cls.api = APIRequestor(**kwargs)
    
    @classmethod
    def get_headers(
        cls,
        **kwargs
    ) -> Dict:

        cls.init_api()
        return cls.api.create_headers(**kwargs)

    @classmethod
    def request(
        cls,
        **kwargs
    ) -> httpx.Response:
        cls.init_api()
        return cls.api._client.request(**kwargs)
    
    @classmethod
    async def async_request(
        cls,
        **kwargs
    ) -> httpx.Response:
        cls.init_api()
        return await cls.api._aclient.request(**kwargs)



import inspect
from typing import Dict
from fileio.utils.helpers import timed, timer

def merge_dicts(x: Dict, y: Dict):
    z = x.copy()
    z.update(y)
    return z

def is_coro_func(obj, func_name: str = None):
    """
    This is probably in the library elsewhere but returns bool
    based on if the function is a coro
    """
    try:
        if inspect.iscoroutinefunction(obj): return True
        if inspect.isawaitable(obj): return True
        if func_name and hasattr(obj, func_name) and inspect.iscoroutinefunction(getattr(obj, func_name)):
            return True
        return bool(hasattr(obj, '__call__') and inspect.iscoroutinefunction(obj.__call__))

    except Exception:
        return False



from __future__ import absolute_import

from async_openai.utils import settings
from async_openai.api import OpenAI

# Allow for submodule calling
# following the old API behavior

Completions = OpenAI.completions
Edits = OpenAI.edits




## Deprecated

from typing import List, Any, Optional, Union, Dict, Type
from async_openai.types import BaseModel
from async_openai.schemas.types.base import BaseResult

__all__ = [
    'EngineRequest',
    'EngineData',
    'EngineResult',
]

class EngineRequest(BaseModel):
    engine_id: Optional[str]

class EngineData(BaseModel):
    id: Optional[str]
    object: Optional[str]
    owner: Optional[str]
    ready: Optional[bool]

class EngineResult(BaseResult):
    data: Optional[Union[EngineData, List[EngineData]]]
    _data_model: Optional[Type[EngineData]] = EngineData
    _request: Optional[EngineRequest] = None